#ifndef __MESH_HPP__
#define __MESH_HPP__

#include "math.hpp"

/**
 * @struct vertex3f
 * @brief Struktura opisujaca wierzchołek w przestrzeni 3d
 */
struct vertex3f
{
    typedef vector3f position_type;
    typedef vector3f normal_type;

    /**
     * @brief Domyslny konstruktor
     * Ustawia wszystkie wektory na ( 0, 0, 0 )
     */
    vertex3f( void ) { }

    /**
     * @brief Konstruktor ustawiajacy wspolrzedne
     * @param pos pozycja wierzchołka w przestrzeni
     * @param n normalna wierzcholka
     */
    explicit vertex3f( const position_type& pos, const normal_type& n  )
        : position( pos ), normal( n )
    {
    }

    ///pozycja wierzchołka
    position_type position;

    ///normalna w wierzchołku
    normal_type normal;
};

/** Deklaracja typu tablicy wierzcholkow */
typedef std::vector< vertex3f > vertexes;

/**
 * @brief Tworzy tablicę wierzchołków płaszczyzny
 * @param width szerokosc plaszczyzny
 * @param height dlugosc plaszczyzny
 */
vertexes createPlane( GLfloat width, GLfloat height )
{
    //rezerwacja pamięci
    vertexes ret;
    ret.reserve( 4 * sizeof( vertexes::value_type ) );

    //wartości pomocnicze
    GLfloat halfWidth = width * 0.5f;
    GLfloat halfHeight = height * 0.5f;

    //lewy górny róg
    ret.push_back(
        vertex3f( vector3f( -halfWidth, .0f, -halfHeight ),
                  vector3f( .0f, 1.0f, .0f ) ) );

    ret.push_back(
        vertex3f( vector3f( -halfWidth, .0f, halfHeight ),
                  vector3f( .0f, 1.0f, .0f ) ) );

    ret.push_back(
        vertex3f( vector3f( halfWidth, .0f, halfHeight ),
                  vector3f( .0f, 1.0f, .0f ) ) );

    ret.push_back(
        vertex3f( vector3f( halfWidth, .0f, -halfHeight ),
                  vector3f( .0f, 1.0f, .0f ) ) );

    return ret;
}

/**
 * @brief przeksztalca podane w parametrach wartosci wspolrzednych
 * w ukladzie sferycznym na wartosci wspolrzednych w ukladzie kartezjanskim
 * @param radius promien sfery
 * @param theta dlugosc geograficzna
 * @param phi szerokosc geograficzna
 */
vector3f spherical_2_car( GLfloat radius, GLfloat theta, GLfloat phi )
{
    //warunki wstępne
    assert( radius > 0 );
    assert( theta >= 0 && theta <= T_PI + EPSILON );
    assert( phi >= 0 && phi <= PI + EPSILON );

    return vector3f(   radius * sin( theta ) * sin( phi )
                     , radius * cos( phi )
                     , radius * sin( phi ) * cos( theta ) );
}

/*
 * @brief Tworzy siatke sfery dla zadanych parametrow.
 * @param radius promien sfery
 * @param slices liczba podzialow dlugosci geograficznej
 * @param segments liczba podzialow szerokosci geograficznej
 */
vertexes createSphere( GLfloat radius, size_t slices, size_t sides )
{
    //warunki wstępne
    assert( radius > 0 );

    vertexes ret;
    ret.reserve( slices * sides * 4 );

    for( size_t i = 0; i < slices; ++i )
    {
        GLfloat dslices = i /
            static_cast< GLfloat >( slices );
        GLfloat dslices_pp = ( i + 1 ) /
            static_cast< GLfloat >( slices );

        GLfloat phi = PI * dslices;
        GLfloat phi_pp = PI * dslices_pp;

        for( size_t j = 0; j < sides; ++j )
        {
            GLfloat dsides = j /
                static_cast< GLfloat >( sides );
            GLfloat dsides_pp = ( j + 1 ) /
                static_cast< GLfloat >( sides );

            GLfloat theta = T_PI * dsides;
            GLfloat theta_pp = T_PI * dsides_pp;

            ret.push_back(
                vertex3f( spherical_2_car( radius, theta, phi ),
                          spherical_2_car( radius
                            , theta, phi ).normalize() ) );

            ret.push_back(
                vertex3f( spherical_2_car( radius, theta, phi_pp ),
                          spherical_2_car( radius
                            , theta, phi_pp ).normalize() ) );

            ret.push_back(
                vertex3f( spherical_2_car( radius, theta_pp, phi_pp ),
                          spherical_2_car( radius
                            , theta_pp, phi_pp ).normalize() ) );

            ret.push_back(
                vertex3f( spherical_2_car( radius, theta_pp, phi ),
                          spherical_2_car( radius
                            , theta_pp, phi ).normalize() ) );
        }
    }

    return ret;
}

/**
 * @brief Tworzy tablice wierzcholkow prostopadłościanu
 * @param width dlugosc boku rownoleglego do osi OX prostopadłoscianu
 * @param height dlugosc boku rownoleglego do osy OZ prostopadłoscianu
 * @param thickness grugosc prostopadloscianu
 */
vertexes createRectangular( GLfloat width, GLfloat height, GLfloat thickness )
{
    vertexes ret;
    //rezerwacja pamięci
    ret.reserve( 6 * 4 * sizeof( vertexes::value_type ) );

    //zapamiętanie wartości pomocniczych
    GLfloat x_min = -width * 0.5f;
    GLfloat y_min = -height * 0.5f;
    GLfloat z_min = -thickness * 0.5f;

    GLfloat x_max = -x_min;
    GLfloat y_max = -y_min;
    GLfloat z_max = -z_min;


    //przednia sciana
    ret.push_back(
        vertex3f( vector3f( x_min, y_max, z_min ),
                  vector3f( .0f, .0f, 1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_min, y_min, z_min ),
                  vector3f( .0f, .0f, 1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_max, y_min, z_min ),
                  vector3f( .0f, .0f, 1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_max, y_max, z_min ),
                  vector3f( .0f, .0f, 1.0f ) ) );

    //tylnia sciana
    ret.push_back(
        vertex3f( vector3f( x_max, y_max, z_min ),
                  vector3f( .0f, .0f, -1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_max, y_min, z_min ),
                  vector3f( .0f, .0f, -1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_min, y_min, z_min ),
                  vector3f( .0f, .0f, -1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_min, y_max, z_min ),
                  vector3f( .0f, .0f, -1.0f ) ) );

    //prawa sciana
    ret.push_back(
        vertex3f( vector3f( x_max, y_max, z_max ),
                  vector3f( 1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_max, y_min, z_max ),
                  vector3f( 1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_max, y_min, z_min ),
                  vector3f( 1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_max, y_max, z_min ),
                  vector3f( 1.0f, .0f, .0f ) ) );

    //lewa sciana
    ret.push_back(
        vertex3f( vector3f( x_min, y_max, z_min ),
                  vector3f( -1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_min, y_min, z_min ),
                  vector3f( -1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_min, y_min, z_max ),
                  vector3f( -1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_min, y_max, z_max ),
                  vector3f( -1.0f, .0f, .0f ) ) );

    //gorna sciana
    ret.push_back(
        vertex3f( vector3f( x_min, y_max, z_min ),
                  vector3f( .0f, 1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_min, y_max, z_max ),
                  vector3f( .0f, 1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_max, y_max, z_max ),
                  vector3f( .0f, 1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_max, y_max, z_min ),
                  vector3f( .0f, 1.0f, .0f ) ) );

    //dolna sciana
    ret.push_back(
        vertex3f( vector3f( x_min, y_min, z_max ),
                  vector3f( .0f, -1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_min, y_min, z_min ),
                  vector3f( .0f, -1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_max, y_min, z_min ),
                  vector3f( .0f, -1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( x_max, y_min, z_max ),
                  vector3f( .0f, -1.0f, .0f ) ) );

    return ret;
}


/**
 * @brief Tworzy tablice wierzcholkow szescianu
 * @param width dlugosc boku szescianu
 */
vertexes createCube( GLfloat width )
{
    vertexes ret;
    //rezerwacja pamięci
    ret.reserve( 6 * 4 * sizeof( vertexes::value_type ) );

    //zapamiętanie wartości pomocniczych
    GLfloat min = -width * 0.5f;
    GLfloat max = -min;

    //przednia sciana
    ret.push_back(
        vertex3f( vector3f( min, max, max ),
                  vector3f( .0f, .0f, 1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( min, min, max ),
                  vector3f( .0f, .0f, 1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, min, max ),
                  vector3f( .0f, .0f, 1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, max, max ),
                  vector3f( .0f, .0f, 1.0f ) ) );

    //tylnia sciana
    ret.push_back(
        vertex3f( vector3f( max, max, min ),
                  vector3f( .0f, .0f, -1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, min, min ),
                  vector3f( .0f, .0f, -1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( min, min, min ),
                  vector3f( .0f, .0f, -1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( min, max, min ),
                  vector3f( .0f, .0f, -1.0f ) ) );

    //prawa sciana
    ret.push_back(
        vertex3f( vector3f( max, max, max ),
                  vector3f( 1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, min, max ),
                  vector3f( 1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, min, min ),
                  vector3f( 1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, max, min ),
                  vector3f( 1.0f, .0f, .0f ) ) );

    //lewa sciana
    ret.push_back(
        vertex3f( vector3f( min, max, min ),
                  vector3f( -1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( min, min, min ),
                  vector3f( -1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( min, min, max ),
                  vector3f( -1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( min, max, max ),
                  vector3f( -1.0f, .0f, .0f ) ) );

    //gorna sciana
    ret.push_back(
        vertex3f( vector3f( min, max, min ),
                  vector3f( .0f, 1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( min, max, max ),
                  vector3f( .0f, 1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, max, max ),
                  vector3f( .0f, 1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, max, min ),
                  vector3f( .0f, 1.0f, .0f ) ) );

    //dolna sciana
    ret.push_back(
        vertex3f( vector3f( min, min, max ),
                  vector3f( .0f, -1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( min, min, min ),
                  vector3f( .0f, -1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, min, min ),
                  vector3f( .0f, -1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, min, max ),
                  vector3f( .0f, -1.0f, .0f ) ) );

    return ret;
}

#endif // __MESH_HPP__
